﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Drawing;

using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;

namespace HeuristicLab.OptimizationExpertSystem {
  [Item("Expert-System Optimizer", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 118)]
  public sealed class ExpertSystemOptimizer : ParameterizedNamedItem, IOptimizer, IStorableContent, INotifyPropertyChanged {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
      private set {
        if (value == null) throw new ArgumentNullException("value");
        if (runs == value) return;
        runs = value;
        OnPropertyChanged("Runs");
      }
    }

    [Storable]
    private ExecutionState executionState;
    public ExecutionState ExecutionState {
      get { return executionState; }
      private set {
        if (executionState != value) {
          executionState = value;
          OnPropertyChanged("ExecutionState");
          OnExecutionStateChanged();
          OnItemImageChanged();
        }
      }
    }

    [Storable]
    private TimeSpan executionTime;
    public TimeSpan ExecutionTime {
      get { return executionTime; }
      private set {
        if (executionTime == value) return;
        executionTime = value;
        OnPropertyChanged("ExecutionTime");
        OnExecutionTimeChanged();
      }
    }

    [Storable]
    private RunCollection knowledgeBase;
    public RunCollection KnowledgeBase {
      get { return knowledgeBase; }
      private set {
        if (value == null) throw new ArgumentNullException("value");
        if (knowledgeBase == value) return;
        knowledgeBase = value;
        OnPropertyChanged("KnowledgeBase");
      }
    }

    [Storable]
    private IProblem problem;
    public IProblem Problem {
      get { return problem; }
      set {
        if (problem == value) return;
        problem = value;
        OnPropertyChanged("Problem");
      }
    }

    [Storable]
    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    public IEnumerable<IOptimizer> NestedOptimizers {
      get { return SuggestedInstances; }
    }

    [StorableConstructor]
    private ExpertSystemOptimizer(bool deserializing) : base(deserializing) { }
    private ExpertSystemOptimizer(ExpertSystemOptimizer original, Cloner cloner)
      : base(original, cloner) {
      runs = cloner.Clone(original.runs);
      executionState = original.executionState;
      executionTime = original.executionTime;
      knowledgeBase = cloner.Clone(original.knowledgeBase);
      problem = cloner.Clone(original.problem);
      suggestedInstances = cloner.Clone(original.suggestedInstances);
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
    }
    public ExpertSystemOptimizer() {
      Runs = new RunCollection();
      KnowledgeBase = new RunCollection();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      ExecutionState = ExecutionState.Stopped;
      ExecutionTime = TimeSpan.Zero;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExpertSystemOptimizer(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
    }

    public void Prepare() {
      Prepare(false);
    }

    public void Prepare(bool clearRuns) {
      if ((ExecutionState != ExecutionState.Prepared) && (ExecutionState != ExecutionState.Paused) && (ExecutionState != ExecutionState.Stopped))
        throw new InvalidOperationException(string.Format("Prepare not allowed in execution state \"{0}\".", ExecutionState));
      if (clearRuns) Runs.Clear();
      ExecutionTime = TimeSpan.Zero;
      ExecutionState = ExecutionState.Prepared;
      OnPrepared();
    }

    public void Start() {
      if ((ExecutionState != ExecutionState.Prepared) && (ExecutionState != ExecutionState.Paused))
        throw new InvalidOperationException(string.Format("Start not allowed in execution state \"{0}\".", ExecutionState));
      ExecutionState = ExecutionState.Started;
      OnStarted();
    }

    public void Pause() {
      if (ExecutionState != ExecutionState.Started)
        throw new InvalidOperationException(string.Format("Pause not allowed in execution state \"{0}\".", ExecutionState));
      ExecutionState = ExecutionState.Paused;
      OnPaused();
    }


    public void Stop() {
      if ((ExecutionState != ExecutionState.Started) && (ExecutionState != ExecutionState.Paused))
        throw new InvalidOperationException(string.Format("Stop not allowed in execution state \"{0}\".", ExecutionState));
      ExecutionState = ExecutionState.Stopped;
      OnStopped();
    }

    public event EventHandler Prepared;
    private void OnPrepared() {
      var handler = Prepared;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler Started;
    private void OnStarted() {
      var handler = Started;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler Paused;
    private void OnPaused() {
      var handler = Paused;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler Stopped;
    private void OnStopped() {
      var handler = Stopped;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler<EventArgs<Exception>> ExceptionOccurred;
    private void OnExceptionOccurred(Exception e) {
      var handler = ExceptionOccurred;
      if (handler != null) handler(this, new EventArgs<Exception>(e));
    }

    public event EventHandler ExecutionStateChanged;
    private void OnExecutionStateChanged() {
      var handler = ExecutionStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler ExecutionTimeChanged;
    private void OnExecutionTimeChanged() {
      var handler = ExecutionTimeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string propertyName) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }
  }
}
